/*
 * Created on 20.02.2003
 *
 */
package de.zebee.mpa;

import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;

import de.zebee.coding.CRC;


/**
 * @author Sebastian Gesemann
 */
public class MP3brfix implements Runnable, JunkHandler {


	public static void main(String [] args) {
		System.out.println("\nMP3brfix (bit-reservoir-fix) v0.95(beta) (c) 2003-2005 by ZeBee\n");
		if ((args==null) || (args.length<1)) {
			System.out.println("Syntax: MP3brfix [options] <file-or-directory>\n");
			System.out.println("If a directory is specified MP3brfix will run on every mp3 it finds");
			System.out.println("in the directory\n");
			System.out.println("Options:");
			System.out.println("-r  scan subdirectories recursivly");
			System.out.println("-f  fix file(s)");
		} else {
			try {
				boolean recursivly = false;
				boolean doFixing = false;
				int paramOfs = 0;
				while (paramOfs<args.length) {
					String str = args[paramOfs];
					if (str.equals("-r")) recursivly = true; else
					if (str.equals("-f")) doFixing = true; else break;
					paramOfs++;
				}
				if (paramOfs>=args.length) {
					System.out.println("missing file/directory parameter!");
					return;
				}
				File file = new File(args[paramOfs++]);
				if (file.isDirectory()) {
					int[] errCount = new int[1];
					int res = runOnDirectory(file,recursivly,doFixing,errCount);
					System.out.println(res+" file(s) processed.");
					System.out.println(errCount[0]+" file(s) had bitres errors.");
				} else {
					runOnFile(file,doFixing);
				}
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
	}

	private static int runOnDirectory(File dir, boolean recursively, boolean fix, int[] errResCount) throws IOException {
		if (!dir.isDirectory()) throw new IllegalArgumentException("directory instead of file expected");
		File[] files = dir.listFiles();
		if (files==null) return 0;
		int fileCount = 0;
		int erroCount = 0;
		int[] temp = new int[1];
		for (int i=0; i<files.length; i++) {
			File file = files[i];
			if (file.isDirectory()) {
				if (recursively) { 
					temp[0] = 0;
					fileCount += runOnDirectory(file,recursively,fix,temp);
					erroCount += temp[0];
				}
			} else {
				String fn = file.getName().toLowerCase();
				if (fn.endsWith(".mp3")) {
					if (runOnFile(file,fix)) {
						erroCount++;
					}
					fileCount++;
				}
			}
		}
		errResCount[0] = erroCount;
		return fileCount;
	}

	private static boolean runOnFile(File file, boolean fix) throws IOException {
		if (file.isDirectory()) throw new IllegalArgumentException("file instead of directory expected");
		InputStream ips = new FileInputStream(file);
		MP3brfix mp3brfix = new MP3brfix(ips,null,true);
		System.out.println("scanning "+file+" ...");
		mp3brfix.run();
		boolean errors = mp3brfix.isStreamErrornous();
		if (errors && fix) {
			File dest = getTempFileFor(file);
			ips = new FileInputStream(file);
			OutputStream ops = new FileOutputStream(dest);
			mp3brfix = new MP3brfix(ips,ops,true);
			System.out.println("fixing "+file+" ...");
			mp3brfix.run();
			String origFileName = file.getAbsolutePath();
			if (file.delete()) {
				if (!dest.renameTo(new File(origFileName))) {
					throw new IOException("Could not rename temp file!");
				}
			} else {
				throw new IOException("Could not delete original file!");
			}
		}
		return errors;
	}

	private static File getTempFileFor(File file) {
		return new File(file.getParent(),file.getName()+".temp");
	}

	private static final int REPORT_MAX_CRC_ERRORS = 20;

	private InputStream ips;
	private OutputStream ops;
	private boolean freeStreams;
	private CRC crc16 = new CRC(CRC.CRC_16);

	private boolean isMPEG1;	// set by run() in the frame loop, used by fixFrame() ...
	private int sideInfoStart;	// "
	private int sideInfoSize;	// "
	private int sideInfoEnd;	// "
	private int byteReservoir;	// "
	private int brErrorCount = 0; 


	public MP3brfix(InputStream ips, OutputStream ops, boolean freeStreams) {
		this.ips = ips;
		this.ops = ops;
		this.freeStreams = freeStreams;
	}

	private static int getSideInfoSize(FrameHeader fHeader) {
		boolean stereo = (fHeader.getChannels() == 2);
		if (fHeader.getMpegID() == FrameHeader.MPEG1_ID) {
			return stereo ? 32 : 17;
		} else {
			return stereo ? 17 : 9;
		}
	}

	public boolean isStreamErrornous() {
		return (brErrorCount > 0);
	}

	public void run() {
		int frameCount = 0;
		brErrorCount = 0;
		int crcErrorCount = 0;
		int bitrateFirst = 0;
		int sumBytes = 0;
		int samplingRateHz = 0;
		int samplesPerFrame = 0;
		int crcFrames = 0;
		boolean modeVBR = false;
		boolean firstFrameIsTagFrame = false;
		try {
			byte[] frameData = new byte[FrameHeader.MAX_MPAFRAME_SIZE];
			MPAFrameParser mpaFP = new MPAFrameParser(ips,(JunkHandler)this);
			FrameHeader fHeader = null;
			int filter = MPAFrameParser.FILTER_LAYER3;
			int brPointer = 0;
			int maxReservoir = 0;
			while (true) {
				try {
					fHeader = mpaFP.getNextFrame(filter,frameData,fHeader);
				} catch (EOFException x) {
					break;
				} 
				sumBytes += fHeader.getFrameSize();
				if (frameCount==0) {
					filter |= MPAFrameParser.getFilterFor(fHeader);
					samplesPerFrame = fHeader.getSamplesPerFrame();
					sideInfoSize = getSideInfoSize(fHeader);
					bitrateFirst = fHeader.getBitrateKBPS();
					samplingRateHz = fHeader.getSamplingrateHZ();
					isMPEG1 = fHeader.getMpegID() == FrameHeader.MPEG1_ID;
					maxReservoir = isMPEG1 ? 511 : 255;
					System.out.println("first frame:\n  "+fHeader);
					firstFrameIsTagFrame = checkXingVbrTag(fHeader,frameData);
				} else {
					if (fHeader.getBitrateKBPS() != bitrateFirst) modeVBR = true;
				}
				if (fHeader.isCrc16used()) {
					crcFrames++;
					if (!checkCRC(fHeader,frameData)) {
						crcErrorCount++;
						if (crcErrorCount<=REPORT_MAX_CRC_ERRORS) {
							System.out.println("ERROR: CRC16 check failed for frame "+frameCount);
							if (crcErrorCount==REPORT_MAX_CRC_ERRORS) {
								System.out.println("INFO:  CRC16 error report now disabled");
							}
						}
					}
					sideInfoStart = 6;
				} else {
					sideInfoStart = 4;
				}
				sideInfoEnd = sideInfoStart + sideInfoSize;
				if (isMPEG1) {
					// 9-bit pointer for MPEG1
					brPointer = ((((int)frameData[sideInfoStart]) & 0xFF) << 1) | ((((int)frameData[sideInfoStart+1]) & 0x80) >> 7);
				} else {
					// 8-bit pointer for MPEG2/2.5
					brPointer = ((int)frameData[sideInfoStart]) & 0xFF;
				}
				if (brPointer > byteReservoir) {
					System.out.println("ERROR: insufficient amount bits in bitreservoir ! (missing "+(brPointer-byteReservoir)+" bytes)");
					if (ops!=null) {
						System.out.println("    -> silencing frame "+frameCount);
					}
					// zeroing "side-info" data for a silence-frame
					// without eliminating main-data bits for the following frame
					fixFrame(fHeader,frameData);
					brErrorCount++;
				}
				if (ops!=null) {
					ops.write(frameData,0,fHeader.getFrameSize());
				}
				byteReservoir = Math.min(maxReservoir,byteReservoir + (fHeader.getFrameSize() - sideInfoEnd));
				if (frameCount==0 && firstFrameIsTagFrame) {
					byteReservoir = 0;
				}
				frameCount++;
			}
		} catch (IOException ioex) {
			throw new RuntimeException(ioex);
		} finally {
			System.out.print("mp3 frames: "+(frameCount - (firstFrameIsTagFrame ? 1 : 0)));
			if (frameCount==0) {
				System.out.println();
			} else {
				if (modeVBR) {
					double avgFrameSize = (sumBytes+0.0) / (frameCount+0.0);
					double bytesPerSec = (avgFrameSize / samplesPerFrame) * samplingRateHz;
					int ckbps = (int) Math.round(bytesPerSec/12.5);
					System.out.println(", VBR ("+(ckbps / 10)+"."+(ckbps % 10)+" kbps)");
				} else {
					System.out.println(", CBR ("+bitrateFirst+" kbps)");
				}
			}
			System.out.println("Bit-Res errors: "+brErrorCount);
			if (crcFrames>0) {
				System.out.println("CRC errors: "+crcErrorCount+"/"+crcFrames);
			} else {
				System.out.println("no CRC protected frames found.");
			}
			if (freeStreams) {
				try {
					if (ips!=null) { ips.close(); ips=null; }
				} catch (IOException e) {} 
				try {
					if (ops!=null) { ops.close(); ops=null; } 
				} catch (IOException e) {} 
			}
			System.out.println("ready.");
		}
	}


	private boolean checkXingVbrTag(FrameHeader fHeader, byte[] frameData) {
		int offset = 4 + getSideInfoSize(fHeader);
		// The offset is correct -- The 'xing standard' does not account for CRC usage !
		boolean xingTag = (frameData[offset] == 0x58) // 'Xing'
		&& (frameData[offset+1] == 0x69)
		&& (frameData[offset+2] == 0x6E)
		&& (frameData[offset+3] == 0x67);
		boolean vbr2Tag = (frameData[offset] == 0x49) // 'Info'
		&& (frameData[offset+1] == 0x6E)
		&& (frameData[offset+2] == 0x66)
		&& (frameData[offset+3] == 0x6F);
		offset += 4;
		if (!xingTag && !vbr2Tag) {
			System.out.println("  no Xing/Info/LAME tag present");
			return false;
		}
		if (xingTag) {
			System.out.println("  Xing-Tag present");
		} else {
			System.out.println("  Info-Tag present");
		}
		System.out.print("  flags:");
		offset += 3;
		int flags = (int)(frameData[offset++]);
		if ((flags & 0xF)==0) {
			System.out.println(" none.");
		} else {
			if ((flags & 1) != 0) {
				System.out.print(" [fc]");
				offset += 4;	// frame-count
			}
			if ((flags & 2) != 0) {
				System.out.print(" [bc]");
				offset += 4;	// byte-count
			}
			if ((flags & 4) != 0) {
				System.out.print(" [toc]");
				offset += 100;	// toc
			}
			if ((flags & 8) != 0) {
				System.out.print(" [vs]");
				offset += 4;	// vbr-scale
			}
			System.out.println();
		}
		if (frameData[offset++] != 0x4C) return true;	// 'LAME' ?
		if (frameData[offset++] != 0x41) return true;
		if (frameData[offset++] != 0x4D) return true;
		if (frameData[offset++] != 0x45) return true;
		System.out.print("  LAME-Tag present  ( LAME ");
		boolean space = false;
		for (int z=0; (frameData[offset+z]!=0) && (z<5); z++) {
			int bite = ((int)frameData[offset+z]) & 0xFF;
			System.out.print((char)bite);
			space = (bite == 0x20);
		}
		offset += 5;
		if (!space) System.out.print(" ");
		System.out.println(")");
		return true;
	}


	private void fixFrame(FrameHeader fHeader, byte[] frameData) {
		for (int offs=4; offs<sideInfoEnd;) {
			frameData[offs++] = 0;
		}
		// pointer to oldest byte in reservoir ...
		if (isMPEG1) {
			// 9-bit pointer for MPEG1
			frameData[sideInfoStart]=(byte)(byteReservoir >> 1);
			frameData[sideInfoStart+1]=(byte)((byteReservoir & 1) << 7);
		} else {
			// 8-bit pointer for MPEG2/2.5
			frameData[sideInfoStart]=(byte)byteReservoir;
		}
		if (sideInfoStart>4) { // CRC checksum comes first
			// update checksum
			int newChecksum = calcCRC16(fHeader,frameData);
			frameData[4] = (byte)(newChecksum >> 8);	// big-endian
			frameData[5] = (byte)(newChecksum & 0xFF);
		}
	}


	private int calcCRC16(FrameHeader fHeader, byte[] frameData) {
		int offset = 2;
		crc16.setRegisterState(0xFFFF);
		crc16.process(frameData[offset++]);
		crc16.process(frameData[offset++]);
		offset+=2; // skip CRC field
		crc16.process(frameData,offset,getSideInfoSize(fHeader));
		return crc16.getRegisterState();
	}


	private boolean checkCRC(FrameHeader fHeader, byte[] frameData) {
		if (!fHeader.isCrc16used()) return true;
		int offset = 4;
		int storedCRC = 0;
		storedCRC = (storedCRC << 8) | (((int)frameData[offset++]) & 0xFF);
		storedCRC = (storedCRC << 8) | (((int)frameData[offset++]) & 0xFF);
		return (storedCRC == calcCRC16(fHeader,frameData));
	}


//-----[ JunkHandler Interface Implementation ]-----

	private static final int JUNK_BUFFER = 0x400;

	private final byte[] junkBuffer = new byte[JUNK_BUFFER];
	private int jbPosition = 0;

	public void write(int bite) throws IOException {
		junkBuffer[jbPosition++] = (byte)bite;
		if (jbPosition==JUNK_BUFFER) flushJunk();
	}
	
	public void endOfJunkBlock() throws IOException {
		flushJunk();
	}

	private void flushJunk() throws IOException {
		if (jbPosition>0) {
			if (ops!=null) {
				ops.write(junkBuffer,0,jbPosition);
			}
			jbPosition = 0;
		}
	}

}
